import numpy as np
import pandas as pd
import pingouin as pg
from scipy.special import factorial
from sklearn.utils import shuffle
from statsmodels.stats import multitest

# ? dictionary with available methods in statsmodels.stats.multitest.multipletests:
# multitest.multitest_methods_names


def apply_pvalue_correction(
    pvalues: np.ndarray, alpha: float = 0.05, method: str = "bonferroni"
) -> tuple[np.ndarray, np.ndarray]:
    """
    Performs p-value correction using the specified method as in
    statsmodels.stats.multitest.multipletests_.

    .. _statsmodels.stats.multitest.multipletests: \
https://www.statsmodels.org/dev/generated/statsmodels.stats.multitest.multipletests.html

    :param numpy.ndarray pvalues: et of p-values of the individual tests.
    :param float alpha: error rate.
    :param str method: method of p-value correction:

        - 'bonferroni' : one-step correction
        - 'sidak' : one-step correction
        - 'holm-sidak' : step down method using Sidak adjustments
        - 'holm' : step-down method using Bonferroni adjustments
        - 'simes-hochberg' : step-up method (independent)
        - 'hommel' : closed method based on Simes tests (non-negative)
        - 'fdr_bh' : Benjamini/Hochberg (non-negative)
        - 'fdr_by' : Benjamini/Yekutieli (negative)
        - 'fdr_tsbh' : two stage fdr correction (non-negative)
        - 'fdr_tsbky' : two stage fdr correction (non-negative)

    :return: Tuple with two `numpy.array`s, boolen for rejecting H0 hypothesis
             and float for adjusted p-value. Can contain missing values if `pvalues`
             contain missing values.

    Example::

        result = apply_pvalue_correction(pvalues, alpha=0.05, method='bonferroni')
    """
    p = np.array(pvalues)
    mask = np.isfinite(p)
    pval_corrected = np.full(p.shape, np.nan)
    _rejected, _pvals_corrected, _, _ = multitest.multipletests(p[mask], alpha, method)
    pval_corrected[mask] = _pvals_corrected
    rejected = np.full(p.shape, np.nan)
    rejected[mask] = _rejected

    return (rejected, pval_corrected)


def apply_pvalue_fdrcorrection(pvalues, alpha=0.05, method="indep"):
    """
    Performs p-value correction for false discovery rate. For more information visit https://www.statsmodels.org/devel/generated/statsmodels.stats.multitest.fdrcorrection.html.

    :param numpy.ndarray pvalues: et of p-values of the individual tests.
    :param float alpha: error rate.
    :param str method: method of p-value correction ('indep', 'negcorr').
    :return: Tuple with two arrays, boolen for rejecting H0 hypothesis and float for adjusted p-value.

    Example::

        result = apply_pvalue_fdrcorrection(pvalues, alpha=0.05, method='indep')
    """
    rejected, padj = multitest.fdrcorrection(pvalues, alpha, method)

    return (rejected, padj)


def apply_pvalue_twostage_fdrcorrection(pvalues, alpha=0.05, method="bh"):
    """
    Iterated two stage linear step-up procedure with estimation of number of true hypotheses. For more information visit https://www.statsmodels.org/dev/generated/statsmodels.stats.multitest.fdrcorrection_twostage.html.

    :param numpy.ndarray pvalues: et of p-values of the individual tests.
    :param float alpha: error rate.
    :param str method: method of p-value correction ('bky', 'bh').
    :return: Tuple with two arrays, boolen for rejecting H0 hypothesis and float for adjusted p-value.

    Example::

        result = apply_pvalue_twostage_fdrcorrection(pvalues, alpha=0.05, method='bh')
    """
    rejected, padj, num_hyp, alpha_stages = multitest.fdrcorrection_twostage(
        pvalues, alpha, method
    )

    return (rejected, padj)


def apply_pvalue_permutation_fdrcorrection(
    df, observed_pvalues, group, alpha=0.05, permutations=50
):
    """
    This function applies multiple hypothesis testing correction using a permutation-based false discovery rate approach.

    :param df: pandas dataframe with samples as rows and features as columns.
    :param oberved_pvalues: pandas Series with p-values calculated on the originally measured data.
    :param str group: name of the column containing group identifiers.
    :param float alpha: error rate. Values velow alpha are considered significant.
    :param int permutations: number of permutations to be applied.
    :return: Pandas dataframe with adjusted p-values and rejected columns.

    Example::

        result = apply_pvalue_permutation_fdrcorrection(df, observed_pvalues, group='group', alpha=0.05, permutations=50)
    """
    i = permutations
    df_index = df.index.values
    df_columns = df.columns.values
    seen = [str([df_index] + df_columns.tolist())]
    rand_pvalues = []
    while i > 0:
        df_index = shuffle(df_index)
        df_columns = shuffle(df_columns)
        df_random = df.reset_index(drop=True)
        df_random.index = df_index
        df_random.index.name = group
        df_random.columns = df_columns
        rand_index = str([df_random.index] + df_columns.tolist())
        if rand_index not in seen:
            seen.append(rand_index)
            df_random = df_random.reset_index()
            for col in df_random.columns.drop(group):
                rand_pvalues.append(
                    calculate_anova(df_random, column=col, group=group)[-1]
                )
            i -= 1
    rand_pvalues = np.array(rand_pvalues)

    qvalues = []
    for i, row in observed_pvalues.to_frame():
        qvalues.append(
            get_counts_permutation_fdr(
                row["pvalue"], rand_pvalues, df["pvalue"], permutations, alpha
            )
            + (i,)
        )

    qvalues = pd.DataFrame(
        qvalues, columns=["padj", "rejected", "identifier"]
    ).set_index("identifier")

    return qvalues


def calculate_anova(df, column, group="group"):
    """
    Calculates one-way ANOVA using pingouin.

    :param df: pandas dataframe with group as rows and protein identifier as column
    :param str column: name of the column in df to run ANOVA on
    :param str group: column with group identifiers
    :return: Tuple with t-statistics and p-value.
    """
    aov_result = pg.anova(data=df, dv=column, between=group)
    df1, df2, t, pvalue = aov_result[["ddof1", "ddof2", "F", "p-unc"]].values.tolist()[
        0
    ]

    return (column, df1, df2, t, pvalue)


def get_counts_permutation_fdr(value, random, observed, n, alpha):
    """
    Calculates local FDR values (q-values) by computing the fraction of accepted hits from the permuted data over accepted hits from the measured data normalized by the total number of permutations.

    :param float value: computed p-value on measured data for a feature.
    :param numpy.ndarray random: p-values computed on the permuted data.
    :param observed: pandas Series with p-values calculated on the originally measured data.
    :param int n: number of permutations to be applied.
    :param float alpha: error rate. Values velow alpha are considered significant.
    :return: Tuple with q-value and boolean for H0 rejected.

    Example::

        result = get_counts_permutation_fdr(value, random, observed, n=250, alpha=0.05)
    """
    a = random[random <= value].shape[0] + 0.0000000000001  # Offset in case of a = 0.0
    b = (observed <= value).sum()
    qvalue = a / b / float(n)

    return (qvalue, qvalue <= alpha)


def get_max_permutations(df, group="group"):
    """
    Get maximum number of permutations according to number of samples.

    :param df: pandas dataframe with samples as rows and protein identifiers as columns
    :param str group: column with group identifiers
    :return: Maximum number of permutations.
    :rtype: int
    """
    num_groups = len(list(df.index))
    num_per_group = df.groupby(group).size().tolist()
    max_perm = factorial(num_groups) / np.prod(factorial(np.array(num_per_group)))

    return max_perm


def correct_pairwise_ttest(df, alpha, correction="fdr_bh"):
    posthoc_df = pd.DataFrame()
    if "group1" in df and "group2" in df and "posthoc pvalue" in df:
        for comparison in df.groupby(["group1", "group2"]).groups:
            index = df.groupby(["group1", "group2"]).groups.get(comparison)
            posthoc_pvalues = df.loc[index, "posthoc pvalue"].tolist()
            _, posthoc_padj = apply_pvalue_correction(
                posthoc_pvalues, alpha=alpha, method=correction
            )
            if posthoc_df.empty:
                posthoc_df = pd.DataFrame(
                    {"index": index, "posthoc padj": posthoc_padj}
                )
            else:
                posthoc_df = posthoc_df.append(
                    pd.DataFrame({"index": index, "posthoc padj": posthoc_padj})
                )
        posthoc_df = posthoc_df.set_index("index")
        df = df.join(posthoc_df)

    return df



"""Run UMAP on DataFrame and return result."""
import pandas as pd
import umap


def run_umap(X_scaled, y, random_state=42) -> pd.DataFrame:
    reducer = umap.UMAP(random_state=random_state, n_jobs=1)
    embedding = reducer.fit_transform(X_scaled)
    embedding = pd.DataFrame(
        embedding, index=X_scaled.index, columns=["UMAP 1", "UMAP 2"]
    ).join(y.astype("category"))
    return embedding



import numpy as np
import pytest

from acore.multiple_testing import apply_pvalue_correction

# benferoni correction
array = [
    0.1,
    0.01,
    0.2,
    0.4,
    np.nan,
    0.5,
]  # 0.04, 0.02, 0.0001]

test_res = [
    (
        "b",
        np.array([0.0, 1.0, 0.0, 0.0, np.nan, 0.0]),
        np.array([0.5, 0.05, 1.0, 1.0, np.nan, 1.0]),
    ),
    (
        "fdr_bh",
        np.array(
            [
                0.0,
                1.0,
                0.0,
                0.0,
                np.nan,
                0.0,
            ]
        ),
        np.array([0.25, 0.05, 0.33333333, 0.5, np.nan, 0.5]),
    ),
]


@pytest.mark.parametrize("method,exp_rejected,exp_pvalues", test_res)
def test_apply_pvalue_correction_alpha_5_percent(method, exp_rejected, exp_pvalues):
    act_rejected, act_pvalues = apply_pvalue_correction(
        array, alpha=0.05, method=method
    )
    np.testing.assert_array_equal(act_rejected, exp_rejected)
    np.testing.assert_array_almost_equal(act_pvalues, exp_pvalues)



import unittest

import numpy as np
import pandas as pd
import pingouin as pg
from scipy import stats

import acore.correlation_analysis as ca


class TestCalculateCorrelations(unittest.TestCase):
    def test_pearson_correlation(self):
        x = np.array([1.5, 0.2, 3.3, 4.34, 5.03])
        y = np.array([2.04, 4.9, 3.6, 0.8, 10.9])
        coefficient, pvalue = ca.calculate_correlations(x, y, method="pearson")
        expected_coefficient, expected_pvalue = stats.pearsonr(x, y)
        self.assertAlmostEqual(coefficient, expected_coefficient)
        self.assertAlmostEqual(pvalue, expected_pvalue)

    def test_spearman_correlation(self):
        x = np.array([1, 2, 3, 4, 5])
        y = np.array([2, 4, 6, 8, 10])
        coefficient, pvalue = ca.calculate_correlations(x, y, method="spearman")
        expected_coefficient, expected_pvalue = stats.spearmanr(x, y)
        self.assertAlmostEqual(coefficient, expected_coefficient)
        self.assertAlmostEqual(pvalue, expected_pvalue)

    def test_calculate_rm_correlation(self):
        # Sample test data
        df = pg.read_dataset("rm_corr")
        x = "pH"
        y = "PacO2"
        subject = "Subject"
        # Expected output
        expected_result = pg.rm_corr(data=df, x=x, y=y, subject=subject)

        # Call the function
        result = ca.calculate_rm_correlation(df, x, y, subject)

        # Compare the results
        self.assertAlmostEqual(result[2], expected_result["r"].values[0])
        self.assertAlmostEqual(result[3], expected_result["pval"].values[0])
        self.assertEqual(result[4], expected_result["dof"].values[0])


def test_corr_lower_triangle():
    df = pd.DataFrame({"A": [1, 2, 3], "B": [4, 5, 7], "C": [6, 8, 9]})
    expected_result = pd.DataFrame(
        {
            "A": {"A": np.nan, "B": 0.9819805060619659, "C": 0.9819805060619656},
            "B": {"A": np.nan, "B": np.nan, "C": 0.9285714285714283},
            "C": {"A": np.nan, "B": np.nan, "C": np.nan},
        }
    )
    result = ca.corr_lower_triangle(df)
    pd.testing.assert_frame_equal(result, expected_result)


if __name__ == "__main__":
    unittest.main()



# %% [markdown]
# (nb_ref_ovarian_data)=
# # Download from journal (ovarian cancer proteome)
# Download the ovarian cancer proteome data from the journal's website. It was
# provided as supplementary data. See the article here:
#
# > Fabian Coscia, Ernst Lengyel, Jaikumar Duraiswamy, Bradley Ashcroft, Michal Bassani-Sternberg, Michael Wierer, Alyssa Johnson, Kristen Wroblewski, Anthony Montag, S. Diane Yamada, Blanca López-Méndez, Jakob Nilsson, Andreas Mund, Matthias Mann, Marion Curtis,
# > Multi-level Proteomics Identifies CT45 as a Chemosensitivity Mediator and Immunotherapy Target in Ovarian Cancer,
# > Cell,
# > Volume 175, Issue 1,
# > 2018,
# >
# > https://doi.org/10.1016/j.cell.2018.08.065.

# %%
# %pip install acore openpyxl

# %%
from pathlib import Path

import pandas as pd

import acore.io

# %% [markdown]
# Specify the proteome file's url

# %%
furl_pg = "https://ars.els-cdn.com/content/image/1-s2.0-S0092867418311668-mmc2.xlsx"

# %% [markdown]
# Load it using an acore function

# %%
local_filename = Path(furl_pg).name
acore.io.download_file(furl_pg, local_filename)

# %% [markdown]
# Open the excel file from the supplementary data given in the article.

# %%
data = pd.read_excel(
    local_filename,
    sheet_name="SupplementaryTable2_PatientProt",
)
data

# %% [markdown]
# We will use the first protein in the a protein group as identifier,
# which we verify to be unique.

# %%
data["first_prot"] = data["Majority protein Ids"].str.split(";").str[0]
data["first_prot"].nunique() == data["Majority protein Ids"].nunique()
data = data.set_index("first_prot")
assert data.index.is_unique
data

# %% [markdown]
# Filter intensity values for patients

# %%
pgs = data.filter(like="Patient")
pgs

# %% [markdown]
# There are two measurements for patient 11 in the data: 11 and 11B. In the methods of
# the paper it is stated:
#
# "We required a minimum peptide ratio count of 1 to report a quantitative readout
# and averaged the results from duplicate measurements of the same sample."
#
# We will do this manually for patient 11 measurements.

# %%
pgs.filter(like="Patient11").describe()

# %%
pgs = pgs.assign(Patient11=lambda df: df.filter(like="Patient11").mean(axis=1)).drop(
    ["Patient11B"], axis=1
)

# %% [markdown]
# Keep the other information of protein groups as additional annotations on protein groups.

# %%
meta_pgs = data.drop(pgs.columns, axis=1).drop("Patient11B", axis=1)
meta_pgs

# %% [markdown]
# View non-numeric columns of protein group metadata.

# %%
meta_pgs.describe(exclude="number")

# %% [markdown]
# Get rid of potential contaminants (marked with a +, so non missing have `NAN`).

# %%
mask = meta_pgs["Potential contaminant"].isna()
pgs = pgs.loc[mask].T
pgs

# %% [markdown]
# ## Patient metadata
# The patient metadata was only provided as a pdf file. We parsed it and
# saved it as a csv file. You can load it for our GitHub repository:

# %%
# ! ToDo: link file after it is available on the main branch

# %% [markdown]
# Done.



# %% [markdown]
# # Exploratory Analysis

# %%
# %pip install acore

# %%
import pandas as pd

import acore.exploratory_analysis as ea

data = pd.DataFrame(
    {
        "group": ["A", "A", "B", "B"],
        "protein1": [1.4, 2.2, 5.3, 4.2],
        "protein2": [5.6, 0.3, 2.1, 8.1],
        "protein3": [9.1, 10.01, 11.2, 12.9],
    }
)

# %% [markdown]
# Show first two principal components of the data.

# %%
result_dfs, annotation = ea.run_pca(
    data, drop_cols=[], annotation_cols=[], group="group", components=2, dropna=True
)

# %% [markdown]
# Show what was computed:

# %%
result_dfs[0]

# %%
result_dfs[1]

# %%
result_dfs[2]

# %%
annotation

# %% [markdown]
# Visualize UMAP low-dimensional embedding of the data.

# %%
result, annotation = ea.run_umap(
    data,
    drop_cols=["sample", "subject"],
    group="group",
    n_neighbors=10,
    min_dist=0.3,
    metric="cosine",
    dropna=True,
)

# %%
result["umap"]

# %%
annotation

# %% [markdown]
# Make sure to check the parameter annotations in the API docs.



import itertools

import numpy as np
import pandas as pd
import utils
from statsmodels.stats.power import FTestAnovaPower


def power_analysis(
    data,
    group="group",
    groups=None,
    alpha=0.05,
    power=0.8,
    dep_var="nobs",
    figure=False,
):
    quantiles = ["25% qtl es", "mean es", "50% qtl es", "75% qtl es"]
    if groups is None:
        groups = data[group].unique().tolist()
    k_groups = len(groups)
    effect_sizes = set()
    for col in data.drop([group], axis=1).columns:
        for g1, g2 in itertools.combinations(groups, 2):
            sample1 = data.loc[data[group] == g1, col].values
            sample2 = data.loc[data[group] == g2, col].values
            eff_size = np.abs(utils.cohens_d(sample1, sample2, ddof=1))
            effect_sizes.add(eff_size)

    summary_eff = []
    if len(effect_sizes):
        effect_sizes = list(effect_sizes)
        summary_eff = [
            np.percentile(effect_sizes, 25),
            np.mean(effect_sizes),
            np.percentile(effect_sizes, 50),
            np.percentile(effect_sizes, 75),
        ]

    analysis = FTestAnovaPower()
    sample_sizes = np.array(range(3, 150))
    power_list = []
    labels = []
    samples = []
    for ii, es in enumerate(summary_eff):
        p = analysis.power(es, sample_sizes, alpha, k_groups)
        labels.extend(["%s = %4.2F" % (quantiles[ii], es)] * len(p))
        power_list.extend(p)
        samples.extend(sample_sizes)

    power_df = pd.DataFrame(
        data=list(zip(power_list, samples, labels)),
        columns=["power", "#samples", "labels"],
    )
    sample_size = analysis.solve_power(
        summary_eff[1], power=power, alpha=alpha, k_groups=k_groups
    )

    return (sample_size, power_df)



from collections.abc import Iterable

import numpy as np
import pandas as pd


def convertToEdgeList(data, cols):
    """
    This function converts a pandas dataframe to an edge list where index becomes the source nodes and columns the target nodes.

    :param data: pandas dataframe.
    :param list cols: names for dataframe columns.
    :return: Pandas dataframe with columns cols.
    """
    data.index.name = None
    edge_list = data.stack().reset_index()
    edge_list.columns = cols

    return edge_list


def check_is_paired(df, subject, group):
    """
    Check if samples are paired.

    :param df: pandas dataframe with samples as rows and protein identifiers as columns (with additional columns 'group', 'sample' and 'subject').
    :param str subject: column with subject identifiers
    :param str group: column with group identifiers
    :return: True if paired samples.
    :rtype: bool
    """
    is_pair = False
    if subject is not None:
        count_subject_groups = df.groupby(subject)[group].count()
        is_pair = (count_subject_groups > 1).all()

    return is_pair


def transform_into_wide_format(data, index, columns, values, extra=[]):
    """
    This function converts a Pandas DataFrame from long to wide format using
    pandas pivot_table() function.

    :param data: long-format Pandas DataFrame
    :param list index: columns that will be converted into the index
    :param str columns: column name whose unique values will become the new column names
    :param str values: column to aggregate
    :param list extra: additional columns to be kept as columns
    :return: Wide-format pandas DataFrame

    Example::

        result = transform_into_wide_format(df, index='index', columns='x', values='y', extra='group')

    """
    df = pd.DataFrame()
    extra_cols = None
    if data is not None:
        df = data.copy()
        if not df.empty:
            if index is None:
                df = df.reset_index()
                index = "index"
            cols = []
            append_to_list(cols, columns)
            append_to_list(cols, values)
            append_to_list(cols, index)
            if len(extra) > 0:
                append_to_list(extra, index)
                extra_cols = df[extra].set_index(index)
            df = df[cols]
            df = df.drop_duplicates()
            if isinstance(index, list):
                df = df.pivot_table(
                    index=index, columns=columns, values=values, aggfunc="first"
                )
            else:
                df = df.pivot(index=index, columns=columns, values=values)
            if extra_cols is not None:
                df = df.join(extra_cols)

            df = df.drop_duplicates()
            df = df.reset_index()

    return df


def transform_into_long_format(data, drop_columns, group, columns=["name", "y"]):
    """
    Converts a Pandas DataDrame from wide to long format using pd.melt()
    function.

    :param data: wide-format Pandas DataFrame
    :param list drop_columns: columns to be deleted
    :param group: column(s) to use as identifier variables
    :type group: str or list
    :param list columns: names to use for the 1)variable column, and for the 2)value column
    :return: Long-format Pandas DataFrame.

    Example::

        result = transform_into_long_format(df, drop_columns=['sample', 'subject'], group='group', columns=['name','y'])
    """
    long_data = pd.DataFrame()
    if data is not None:
        data = data.drop(drop_columns, axis=1)

        long_data = pd.melt(
            data, id_vars=group, var_name=columns[0], value_name=columns[1]
        )
        long_data = long_data.set_index(group)
        long_data.columns = columns

    return long_data


def remove_group(data):
    """
    Removes column with label 'group'.

    :param data: pandas dataframe with one column labelled 'group'
    :return: Pandas dataframe

    Example::

        result = remove_group(data)
    """
    data.drop(["group"], axis=1)
    return data


def calculate_fold_change(df, condition1, condition2):
    """
    Calculates fold-changes between two groups for all proteins in a dataframe.

    :param df: pandas dataframe with samples as rows and protein identifiers as columns.
    :param str condition1: identifier of first group.
    :param str condition2: identifier of second group.
    :return: Numpy array.

    Example::

        result = calculate_fold_change(data, 'group1', 'group2')
    """
    group1 = df[condition1]
    group2 = df[condition2]

    if isinstance(group1, np.float64):
        group1 = np.array(group1)
    else:
        group1 = group1.values
    if isinstance(group2, np.float64):
        group2 = np.array(group2)
    else:
        group2 = group2.values

    if np.isnan(group1).all() or np.isnan(group2).all():
        fold_change = np.nan
    else:
        fold_change = np.nanmedian(group1) - np.nanmedian(group2)

    return fold_change


def pooled_standard_deviation(sample1, sample2, ddof):
    """
    Calculates the pooled standard deviation.
    For more information visit https://www.hackdeploy.com/learn-what-is-statistical-power-with-python/.

    :param array sample1: numpy array with values for first group
    :param array sample2: numpy array with values for second group
    :param int ddof: degrees of freedom
    """
    # calculate the sample size
    n1, n2 = len(sample1), len(sample2)
    # calculate the variances
    var1, var2 = np.var(sample1, ddof=1), np.var(sample2, ddof=ddof)
    # calculate the pooled standard deviation
    numerator = ((n1 - 1) * var1) + ((n2 - 1) * var2)
    denominator = n1 + n2 - 2
    return np.sqrt(numerator / denominator)


def cohens_d(sample1, sample2, ddof):
    """
    Calculates Cohen's d effect size based on the distance between two means, measured in standard deviations.
    For more information visit https://www.hackdeploy.com/learn-what-is-statistical-power-with-python/.

    :param array sample1: numpy array with values for first group
    :param array sample2: numpy array with values for second group
    :param int ddof: degrees of freedom
    """
    u1, u2 = np.mean(sample1), np.mean(sample2)
    s_pooled = pooled_standard_deviation(sample1, sample2, ddof)

    return (u1 - u2) / s_pooled


def hedges_g(df, condition1, condition2, ddof=0):
    """
    Calculates Hedges’ g effect size (more accurate for sample sizes below 20 than Cohen's d).
    For more information visit https://docs.scipy.org/doc/numpy/reference/generated/numpy.nanstd.html.

    :param df: pandas dataframe with samples as rows and protein identifiers as columns.
    :param str condition1: identifier of first group.
    :param str condition2: identifier of second group.
    :param int ddof: means Delta Degrees of Freedom.
    :return: Numpy array.

    Example::

        result = hedges_g(data, 'group1', 'group2', ddof=0)
    """
    group1 = df[condition1]
    group2 = df[condition2]

    if isinstance(group1, np.float64):
        group1 = np.array(group1)
    else:
        group1 = group1.values
    if isinstance(group2, np.float64):
        group2 = np.array(group2)
    else:
        group2 = group2.values

    ng1 = group1.size
    ng2 = group2.size
    # dof = ng1 + ng2 - 2
    if np.isnan(group1).all() or np.isnan(group2).all():
        g = np.nan
    else:
        meang1 = np.nanmean(group1)
        meang2 = np.nanmean(group2)
        sdpooled = np.nanstd(np.concatenate([group1, group2]), ddof=ddof)

        # Correct bias small sample size
        if ng1 + ng2 < 50:
            g = (
                ((meang1 - meang2) / sdpooled)
                * ((ng1 + ng2 - 3) / (ng1 + ng2 - 2.25))
                * np.sqrt((ng1 + ng2 - 2) / (ng1 + ng2))
            )
        else:
            g = (meang1 - meang2) / sdpooled

    return g


def unit_vector(vector):
    """
    Returns the unit vector of the vector.
    :param tuple vector: vector
    :return tuple unit_vector: unit vector
    """
    return vector / np.linalg.norm(vector)


def flatten(t, my_list=[]):
    """
    Code from: https://gist.github.com/shaxbee/0ada767debf9eefbdb6e
    Acknowledgements: Zbigniew Mandziejewicz (shaxbee)
    Generator flattening the structure

    >>> list(flatten([2, [2, (4, 5, [7], [2, [6, 2, 6, [6], 4]], 6)]]))
    [2, 2, 4, 5, 7, 2, 6, 2, 6, 6, 4, 6]
    """

    for x in t:
        if not isinstance(x, Iterable) or isinstance(x, str):
            my_list.append(x)
        else:
            flatten(x, my_list)

    return my_list


def angle_between(v1, v2):
    """
    Returns the angle in radians between vectors 'v1' and 'v2'

    :param tuple v1: vector 1
    :param tuple v2: vector 2
    :return float angle: angle between two vectors in radians

    Example::
        angle = angle_between((1, 0, 0), (0, 1, 0))
    """
    v1_u = unit_vector(v1)
    v2_u = unit_vector(v2)
    return np.arccos(np.clip(np.dot(v1_u, v2_u), -1.0, 1.0))


def append_to_list(mylist, myappend):
    if isinstance(myappend, list):
        mylist.extend(myappend)
    else:
        mylist.append(myappend)


def generator_to_dict(genvar):
    dictvar = {}
    for i, gen in enumerate(genvar):
        dictvar.update({n: i for n in gen})

    return dictvar



"""Uniprot ID mapping using Python.

Source: https://www.uniprot.org/help/id_mapping
"""

import json
import re
import time
import zlib
from urllib.parse import parse_qs, urlencode, urlparse
from xml.etree import ElementTree

import pandas as pd
import requests
from requests.adapters import HTTPAdapter, Retry

POLLING_INTERVAL = 3
API_URL = "https://rest.uniprot.org"


retries = Retry(total=5, backoff_factor=0.25, status_forcelist=[500, 502, 503, 504])
session = requests.Session()
session.mount("https://", HTTPAdapter(max_retries=retries))


def check_response(response):
    try:
        response.raise_for_status()
    except requests.HTTPError:
        print(response.json())
        raise


def submit_id_mapping(from_db, to_db, ids):
    request = requests.post(
        f"{API_URL}/idmapping/run",
        data={"from": from_db, "to": to_db, "ids": ",".join(ids)},
    )
    check_response(request)
    return request.json()["jobId"]


def get_next_link(headers):
    re_next_link = re.compile(r'<(.+)>; rel="next"')
    if "Link" in headers:
        match = re_next_link.match(headers["Link"])
        if match:
            return match.group(1)


def check_id_mapping_results_ready(job_id):
    while True:
        request = session.get(f"{API_URL}/idmapping/status/{job_id}")
        check_response(request)
        j = request.json()
        if "jobStatus" in j:
            if j["jobStatus"] in ("NEW", "RUNNING"):
                print(f"Retrying in {POLLING_INTERVAL}s")
                time.sleep(POLLING_INTERVAL)
            else:
                raise Exception(j["jobStatus"])
        else:
            return bool(j["results"] or j["failedIds"])


def get_batch(batch_response, file_format, compressed):
    batch_url = get_next_link(batch_response.headers)
    while batch_url:
        batch_response = session.get(batch_url)
        batch_response.raise_for_status()
        yield decode_results(batch_response, file_format, compressed)
        batch_url = get_next_link(batch_response.headers)


def combine_batches(all_results, batch_results, file_format):
    if file_format == "json":
        for key in ("results", "failedIds"):
            if key in batch_results and batch_results[key]:
                all_results[key] += batch_results[key]
    elif file_format == "tsv":
        return all_results + batch_results[1:]
    else:
        return all_results + batch_results
    return all_results


def get_id_mapping_results_link(job_id):
    url = f"{API_URL}/idmapping/details/{job_id}"
    request = session.get(url)
    check_response(request)
    return request.json()["redirectURL"]


def decode_results(response, file_format, compressed):
    if compressed:
        decompressed = zlib.decompress(response.content, 16 + zlib.MAX_WBITS)
        if file_format == "json":
            j = json.loads(decompressed.decode("utf-8"))
            return j
        elif file_format == "tsv":
            return [line for line in decompressed.decode("utf-8").split("\n") if line]
        elif file_format == "xlsx":
            return [decompressed]
        elif file_format == "xml":
            return [decompressed.decode("utf-8")]
        else:
            return decompressed.decode("utf-8")
    elif file_format == "json":
        return response.json()
    elif file_format == "tsv":
        return [line for line in response.text.split("\n") if line]
    elif file_format == "xlsx":
        return [response.content]
    elif file_format == "xml":
        return [response.text]
    return response.text


def get_xml_namespace(element):
    m = re.match(r"\{(.*)\}", element.tag)
    return m.groups()[0] if m else ""


def merge_xml_results(xml_results):
    merged_root = ElementTree.fromstring(xml_results[0])
    for result in xml_results[1:]:
        root = ElementTree.fromstring(result)
        for child in root.findall("{http://uniprot.org/uniprot}entry"):
            merged_root.insert(-1, child)
    ElementTree.register_namespace("", get_xml_namespace(merged_root[0]))
    return ElementTree.tostring(merged_root, encoding="utf-8", xml_declaration=True)


def print_progress_batches(batch_index, size, total):
    n_fetched = min((batch_index + 1) * size, total)
    print(f"Fetched: {n_fetched} / {total}")


def get_id_mapping_results_search(url):
    parsed = urlparse(url)
    query = parse_qs(parsed.query)
    file_format = query["format"][0] if "format" in query else "json"
    if "size" in query:
        size = int(query["size"][0])
    else:
        size = 500
        query["size"] = size
    compressed = (
        query["compressed"][0].lower() == "true" if "compressed" in query else False
    )
    parsed = parsed._replace(query=urlencode(query, doseq=True))
    url = parsed.geturl()
    request = session.get(url)
    check_response(request)
    results = decode_results(request, file_format, compressed)
    total = int(request.headers["x-total-results"])
    print_progress_batches(0, size, total)
    for i, batch in enumerate(get_batch(request, file_format, compressed), 1):
        results = combine_batches(results, batch, file_format)
        print_progress_batches(i, size, total)
    if file_format == "xml":
        return merge_xml_results(results)
    return results


def get_id_mapping_results_stream(url):
    if "/stream/" not in url:
        url = url.replace("/results/", "/results/stream/")
    request = session.get(url)
    check_response(request)
    parsed = urlparse(url)
    query = parse_qs(parsed.query)
    file_format = query["format"][0] if "format" in query else "json"
    compressed = (
        query["compressed"][0].lower() == "true" if "compressed" in query else False
    )
    return decode_results(request, file_format, compressed)




if __name__ == "__main__":
    # id mapping is used to create a link to a query (you can see the json in the browser)
    # UniProtKB is the knowleadgebase integrating all kind of other databases
    import pandas as pd

    job_id = submit_id_mapping(
        from_db="UniProtKB_AC-ID", to_db="UniProtKB", ids=["P05067", "P12345"]
    )

    if check_id_mapping_results_ready(job_id):
        link = get_id_mapping_results_link(job_id)
        # add fields to the link to get more information
        # From and Entry (accession) are the same.
        results = get_id_mapping_results_search(
            link + "?fields=accession,go_p,go_c,go_f&format=tsv"
        )
        # see the available fields you can add
        # https://www.uniprot.org/help/return_fields
        # and the available formats: json and tsv (for most endpoints)
        # https://www.uniprot.org/help/api_queries#tips
    print(results)
    header = results.pop(0).split("\t")
    results = [line.split("\t") for line in results]
    df = pd.DataFrame(results, columns=header)
    # result from:
    # df.to_dict(orient="records")
    records = [
        {
            "From": "P05067",
            "Gene Ontology (biological process)": (
                "adult locomotory behavior [GO:0008344]; amyloid fibril formation"
                " [GO:1990000]; astrocyte activation [GO:0048143]; astrocyte activation"
                " involved in immune response [GO:0002265]; axo-dendritic transport"
                " [GO:0008088]; axon midline choice point recognition [GO:0016199];"
                " axonogenesis [GO:0007409]; cell adhesion [GO:0007155]; cellular"
                " response to amyloid-beta [GO:1904646]; central nervous system"
                " development [GO:0007417]; cholesterol metabolic process [GO:0008203];"
                " cognition [GO:0050890]; collateral sprouting in absence of injury"
                " [GO:0048669]; cytosolic mRNA polyadenylation [GO:0180011]; dendrite"
                " development [GO:0016358]; endocytosis [GO:0006897]; extracellular"
                " matrix organization [GO:0030198]; forebrain development [GO:0030900];"
                " G2/M transition of mitotic cell cycle [GO:0000086]; intracellular"
                " copper ion homeostasis [GO:0006878]; ionotropic glutamate receptor"
                " signaling pathway [GO:0035235]; learning [GO:0007612]; learning or"
                " memory [GO:0007611]; locomotory behavior [GO:0007626]; mating"
                " behavior [GO:0007617]; microglia development [GO:0014005]; microglial"
                " cell activation [GO:0001774]; modulation of excitatory postsynaptic"
                " potential [GO:0098815]; negative regulation of cell population"
                " proliferation [GO:0008285]; negative regulation of gene expression"
                " [GO:0010629]; negative regulation of long-term synaptic potentiation"
                " [GO:1900272]; negative regulation of neuron differentiation"
                " [GO:0045665]; neuromuscular process controlling balance [GO:0050885];"
                " neuron apoptotic process [GO:0051402]; neuron cellular homeostasis"
                " [GO:0070050]; neuron projection development [GO:0031175]; neuron"
                " projection maintenance [GO:1990535]; neuron remodeling [GO:0016322];"
                " NMDA selective glutamate receptor signaling pathway [GO:0098989];"
                " Notch signaling pathway [GO:0007219]; positive regulation of amyloid"
                " fibril formation [GO:1905908]; positive regulation of"
                " calcium-mediated signaling [GO:0050850]; positive regulation of"
                " chemokine production [GO:0032722]; positive regulation of ERK1 and"
                " ERK2 cascade [GO:0070374]; positive regulation of G2/M transition of"
                " mitotic cell cycle [GO:0010971]; positive regulation of gene"
                " expression [GO:0010628]; positive regulation of glycolytic process"
                " [GO:0045821]; positive regulation of inflammatory response"
                " [GO:0050729]; positive regulation of interleukin-1 beta production"
                " [GO:0032731]; positive regulation of interleukin-6 production"
                " [GO:0032755]; positive regulation of JNK cascade [GO:0046330];"
                " positive regulation of long-term synaptic potentiation [GO:1900273];"
                " positive regulation of mitotic cell cycle [GO:0045931]; positive"
                " regulation of non-canonical NF-kappaB signal transduction"
                " [GO:1901224]; positive regulation of peptidyl-serine phosphorylation"
                " [GO:0033138]; positive regulation of peptidyl-threonine"
                " phosphorylation [GO:0010800]; positive regulation of protein"
                " metabolic process [GO:0051247]; positive regulation of protein"
                " phosphorylation [GO:0001934]; positive regulation of T cell migration"
                " [GO:2000406]; positive regulation of transcription by RNA polymerase"
                " II [GO:0045944]; positive regulation of tumor necrosis factor"
                " production [GO:0032760]; regulation of gene expression [GO:0010468];"
                " regulation of long-term neuronal synaptic plasticity [GO:0048169];"
                " regulation of multicellular organism growth [GO:0040014]; regulation"
                " of peptidyl-tyrosine phosphorylation [GO:0050730]; regulation of"
                " presynapse assembly [GO:1905606]; regulation of spontaneous synaptic"
                " transmission [GO:0150003]; regulation of synapse structure or"
                " activity [GO:0050803]; regulation of translation [GO:0006417];"
                " regulation of Wnt signaling pathway [GO:0030111]; response to"
                " interleukin-1 [GO:0070555]; response to oxidative stress"
                " [GO:0006979]; smooth endoplasmic reticulum calcium ion homeostasis"
                " [GO:0051563]; suckling behavior [GO:0001967]; synapse organization"
                " [GO:0050808]; synaptic assembly at neuromuscular junction"
                " [GO:0051124]; visual learning [GO:0008542]"
            ),
            "Gene Ontology (cellular component)": (
                "apical part of cell [GO:0045177]; axon [GO:0030424]; cell surface"
                " [GO:0009986]; cell-cell junction [GO:0005911]; ciliary rootlet"
                " [GO:0035253]; clathrin-coated pit [GO:0005905]; COPII-coated ER to"
                " Golgi transport vesicle [GO:0030134]; cytoplasm [GO:0005737]; cytosol"
                " [GO:0005829]; dendrite [GO:0030425]; dendritic shaft [GO:0043198];"
                " dendritic spine [GO:0043197]; early endosome [GO:0005769];"
                " endoplasmic reticulum [GO:0005783]; endoplasmic reticulum lumen"
                " [GO:0005788]; endosome [GO:0005768]; endosome lumen [GO:0031904];"
                " extracellular exosome [GO:0070062]; extracellular region"
                " [GO:0005576]; extracellular space [GO:0005615]; Golgi apparatus"
                " [GO:0005794]; Golgi lumen [GO:0005796]; Golgi-associated vesicle"
                " [GO:0005798]; growth cone [GO:0030426]; membrane [GO:0016020];"
                " membrane raft [GO:0045121]; mitochondrial inner membrane"
                " [GO:0005743]; neuromuscular junction [GO:0031594]; nuclear envelope"
                " lumen [GO:0005641]; perikaryon [GO:0043204]; perinuclear region of"
                " cytoplasm [GO:0048471]; plasma membrane [GO:0005886]; platelet alpha"
                " granule lumen [GO:0031093]; presynaptic active zone [GO:0048786];"
                " receptor complex [GO:0043235]; recycling endosome [GO:0055037];"
                " smooth endoplasmic reticulum [GO:0005790]; spindle midzone"
                " [GO:0051233]; synapse [GO:0045202]; synaptic vesicle [GO:0008021];"
                " trans-Golgi network membrane [GO:0032588]"
            ),
            "Gene Ontology (molecular function)": (
                "DNA binding [GO:0003677]; enzyme binding [GO:0019899]; heparin binding"
                " [GO:0008201]; identical protein binding [GO:0042802]; protein"
                " serine/threonine kinase binding [GO:0120283]; PTB domain binding"
                " [GO:0051425]; receptor ligand activity [GO:0048018]; RNA polymerase"
                " II cis-regulatory region sequence-specific DNA binding [GO:0000978];"
                " serine-type endopeptidase inhibitor activity [GO:0004867]; signaling"
                " receptor activator activity [GO:0030546]; signaling receptor binding"
                " [GO:0005102]; transition metal ion binding [GO:0046914]"
            ),
        },
        {
            "From": "P12345",
            "Gene Ontology (biological process)": (
                "2-oxoglutarate metabolic process [GO:0006103]; aspartate catabolic"
                " process [GO:0006533]; aspartate metabolic process [GO:0006531];"
                " glutamate metabolic process [GO:0006536]; lipid transport"
                " [GO:0006869]; protein folding [GO:0006457]"
            ),
            "Gene Ontology (cellular component)": (
                "mitochondrial matrix [GO:0005759]; mitochondrion [GO:0005739]; plasma"
                " membrane [GO:0005886]"
            ),
            "Gene Ontology (molecular function)": (
                "kynurenine-oxoglutarate transaminase activity [GO:0016212];"
                " L-aspartate:2-oxoglutarate aminotransferase activity [GO:0004069];"
                " pyridoxal phosphate binding [GO:0030170]"
            ),
        },
    ]



import ftplib
from pathlib import Path


def download_from_ftp(
    ftp_url: str, user: str, password: str, to: str, file_name
) -> str:
    """Download a file from an FTP server."""
    to = Path(to)
    to.mkdir(parents=True, exist_ok=True)
    try:
        domain = ftp_url.split("/")[2]
        ftp_file = "/".join(ftp_url.split("/")[3:])
        with ftplib.FTP(domain) as ftp:
            ftp.login(user=user, passwd=password)
            with open(to / file_name, "wb") as fp:
                ftp.retrbinary("RETR " + ftp_file, fp.write)
    except ftplib.error_reply as err:
        raise ftplib.error_reply(
            "Exception raised when an unexpected reply is received from the server."
            f" {err}.\nURL:{ftp_url}"
        ) from err
    except ftplib.error_temp as err:
        raise ftplib.error_temp(
            "Exception raised when an error code signifying a temporary error."
            f" {err}.\nURL:{ftp_url}"
        ) from err
    except ftplib.error_perm as err:
        raise ftplib.error_perm(
            "Exception raised when an error code signifying a permanent error."
            f" {err}.\nURL:{ftp_url}"
        ) from err
    except ftplib.error_proto as err:
        raise ftplib.error_proto(
            "Exception raised when a reply is received from the server that does not"
            " fit the response specifications of the File Transfer Protocol."
            f" {err}.\nURL:{ftp_url}"
        ) from err



# %% [markdown]
# # Enrichment analysis
# requires
# - some cluster of proteins/genes (e.g. up- and downregulated proteins/genes)
# - functional annotations, i.e. a category summarizing a set of proteins/genes.
#
# You can start with watching Lars Juhl Jensen's brief introduction to enrichment analysis
# on [youtube](https://www.youtube.com/watch?v=2NC1QOXmc5o).
#
# Here we use as example data from an ovarian cancer dataset:
# [PXD010372](https://github.com/Multiomics-Analytics-Group/acore/tree/main/example_data/PXD010372)
#
# First make sure you have the required packages installed:


# %% tags=["hide-output"]
# %pip install acore vuecore 'plotly<6'

# %%
from pathlib import Path

import dsp_pandas
import pandas as pd

import acore
import acore.differential_regulation
import acore.enrichment_analysis

dsp_pandas.format.set_pandas_options(max_colwidth=60)

# %% [markdown]
# Parameters of this notebook

# %% tags=["parameters"]
base_path: str = (
    "https://raw.githubusercontent.com/Multiomics-Analytics-Group/acore/refs/heads/main/"
    "example_data/PXD010372/processed"
)
omics: str = f"{base_path}/omics.csv"
meta_pgs: str = f"{base_path}/meta_pgs.csv"
meta: str = f"{base_path}/meta_patients.csv"
features_to_sample: int = 100

# %% [markdown]
# # Load processed data
# from our repository. See details on obtaining the data under the example data section on
# [this page](nb_ref_ovarian_data)

# %%
df_omics = pd.read_csv(omics, index_col=0)
df_meta_pgs = pd.read_csv(meta_pgs, index_col=0)
df_meta = pd.read_csv(meta, index_col=0)
df_omics

# %%
ax = (
    df_omics.notna()
    .sum()
    .sort_values(ascending=True)
    .plot(xlabel="Protein groups", ylabel="Number of non-NaN values (samples)")
)

# %% [markdown]
# Keep only features with a certain amount of non-NaN values and select 100 of these
# for illustration. Add always four which were differently regulated in the ANOVA using all
# the protein groups.

# %%
idx_always_included = ["Q5HYN5", "P39059", "O43432", "O43175"]

# %% tags=["hide-input"]
df_omics = (
    df_omics
    # .dropna(axis=1)
    .drop(idx_always_included, axis=1)
    .dropna(thresh=18, axis=1)
    .sample(
        features_to_sample - len(idx_always_included),
        axis=1,
        random_state=42,
    )
    .join(df_omics[idx_always_included])
)
df_omics

# %% [markdown]
# And we have the following patient metadata, from which we will use the `Status` column as
# our dependent variable and the `PlatinumValue` as a covariate.

# %%
df_meta


# %% [markdown]
# # ANOVA: Compute up and downregulated genes
# These will be used to find enrichments in the set of both up and downregulated genes.

# %%
group = "Status"
diff_reg = acore.differential_regulation.run_anova(
    df_omics.join(df_meta[[group]]),
    drop_cols=[],
    subject=None,
    group=group,
)
diff_reg.describe(exclude=["float"])

# %%
diff_reg["rejected"] = diff_reg["rejected"].astype(bool)  # ! needs to be fixed in anova
diff_reg.query("rejected")

# %% [markdown]
# # Download functional annotations, here pathways, for the protein groups
# in our selection of the dataset.

# %%
from acore.io.uniprot import fetch_annotations, process_annotations

fname_annotations = f"downloaded/annotations_{features_to_sample}.csv"
fname = Path(fname_annotations)
try:
    annotations = pd.read_csv(fname, index_col=0)
    print(f"Loaded annotations from {fname}")
except FileNotFoundError:
    print(f"Fetching annotations for {df_omics.columns.size} UniProt IDs.")
    FIELDS = "go_p,go_c,go_f"
    annotations = fetch_annotations(df_omics.columns, fields=FIELDS)
    annotations = process_annotations(annotations, fields=FIELDS)
    # cache the annotations
    fname.parent.mkdir(exist_ok=True, parents=True)
    annotations.to_csv(fname, index=True)

annotations

# %% [markdown]
# See how many protein groups are associated with each annotation. We observe that most
# functional annotations are associated only to a single protein group in our dataset.

# %% tags=["hide-input"]
s_count_pg_per_annotation = (
    annotations.groupby("annotation").size().value_counts().sort_index()
)
_ = s_count_pg_per_annotation.plot(
    kind="bar",
    xlabel="Number of protein groups associated with annotation",
    ylabel="Number of annotations",
)
s_count_pg_per_annotation.to_frame("number of annotations").rename_axis(
    "N protein groups"
).T

# %%
annotations.groupby("annotation").size().value_counts(ascending=False)

# %% [markdown]
# # Enrichment analysis
# Is done separately for up- and downregulated genes as it's assumed that biological
# processes are regulated in one direction.

# %% tags=["hide-input"]
diff_reg.query("rejected")[
    [
        "identifier",
        "group1",
        "group2",
        "pvalue",
        "padj",
        "rejected",
        "log2FC",
        "FC",
    ]
].sort_values("log2FC")

# %% [markdown]
# Running the enrichment analysis for the up- and down regulated protein groups
# separately with the default settings of the function, i.e. a log2 fold change cutoff
# of 1 and at least 2 protein groups detected in the set of proteins
# defining the functional annotation.

# %%
ret = acore.enrichment_analysis.run_up_down_regulation_enrichment(
    regulation_data=diff_reg,
    annotation=annotations,
    min_detected_in_set=2,
    lfc_cutoff=1,
)
ret

# %% [markdown]
# we can decrease the cutoff for the log2 fold change to 0.5 and see that we retain
# more annotations.

# %%
ret = acore.enrichment_analysis.run_up_down_regulation_enrichment(
    regulation_data=diff_reg,
    annotation=annotations,
    min_detected_in_set=2,
    lfc_cutoff=0.5,  # ! the default is 1
)
ret

# %% [markdown]
# And even more if we do not restrict the analysis of finding at least two proteins
# of a functional set in our data set (i.e. we only need to find one match from the set).

# %%
ret = acore.enrichment_analysis.run_up_down_regulation_enrichment(
    regulation_data=diff_reg,
    annotation=annotations,
    min_detected_in_set=1,
    lfc_cutoff=0.5,  # ! the default is 1
)
ret

# %% [markdown]
# ## Site specific enrichment analysis

# %% [markdown]
# The basic example uses a modified peptide sequence to
# demonstrate the enrichment analysis.
# > TODO: The example on how to do that needs a PTM focused dataset.
# The details of how site specific enrichment analysis is done will depend on the
# dataset and the question at hand.
#
# If the identifiers contain PTMs this information is removed to match it to the annotation
# using a regular expression (in the function). For example:

# %%
import re

regex = "(\\w+~.+)_\\w\\d+\\-\\w+"
identifier_ckg = "gnd~P00350_T10-WW"
match = re.search(regex, identifier_ckg)
match.group(1)

# %%
# ToDo: Add example for site specific enrichment analysis
# acore.enrichment_analysis.run_up_down_regulation_enrichment

# %% [markdown]
# # Single sample GSEA (ssGSEA)
# Run a gene set enrichment analysis (GSEA) for each sample,
# see [article](https://www.nature.com/articles/nature08460#Sec3) and
# the package [`gseapy`](https://gseapy.readthedocs.io/en/latest/run.html#gseapy.ssgsea)
# for more details.

# %%
enrichtments = acore.enrichment_analysis.run_ssgsea(
    data=df_omics,
    annotation=annotations,
    min_size=1,
)
enrichtments

# %%
enrichtments.iloc[0].to_dict()

# %%
ax = enrichtments["NES"].plot.hist()

# %% [markdown]
# The normalised enrichment score (NES) can be used in a PCA plot to see if the samples
# cluster according to the enrichment of the gene sets.

# %%
nes = enrichtments.set_index("Term", append=True).unstack()["NES"].convert_dtypes()
nes

# %%
import acore.exploratory_analysis as ea

pca_result, pca_annotation = ea.run_pca(
    data=nes.join(df_meta[[group]]),
    drop_cols=[],
    annotation_cols=[],
    group=group,
    components=2,
    dropna=False,
)
resultDf, loadings, var_exp = pca_result
resultDf

# %% [markdown]
# The loadings show how the variables are correlated with the principal components.

# %%
loadings

# %% [markdown]
# We will plot both on the sample plot (samples on the first two principal components and
# loadings of variables). We use the
# [`vuecore` package](https://github.com/Multiomics-Analytics-Group/vuecore)
# for this, which is also developed by the Multiomics Analytics Group.

# %%
import plotly.graph_objects as go
from vuecore import viz

args = {"factor": 2, "loadings": 1}  # increase number of loadings or scaling factor
# #! pca_results has three items, but docstring requests only two -> double check
figure = viz.get_pca_plot(data=pca_result, identifier="PCA enrichment", args=args)
figure = go.Figure(data=figure["data"], layout=figure["layout"])
figure.show()

# %% [markdown]
# # Compare two distributions - KS test
#
# The Kolmogorov-Smirnov test is a non-parametric test that compares two distributions.
# - we compare the distributions of the two differently upregulated protein groups
# This is not the best example for comparing distributions, but it shows how to use the
# KS test.

# %%
# plot two histograms of intensity values here
sel_pgs = ["O43175", "P39059"]
view = df_omics[sel_pgs].sub(df_omics[sel_pgs].mean())
ax = view.plot.hist(bins=20, alpha=0.5)

# %% [markdown]
# Let us compare the two centered distributions using the KS test.

# %%
acore.enrichment_analysis.run_kolmogorov_smirnov(view[sel_pgs[0]], view[sel_pgs[1]])

# %% [markdown]
# The result suggests that the two distributions are from the same distribution.



import unittest
import pandas as pd
import pingouin as pg
import acore.differential_regulation as dr


class TestCalculateTtest(unittest.TestCase):
    def setUp(self):
        self.data = {
            "subject": [1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4],
            "sample": [
                "S1",
                "S2",
                "S3",
                "S4",
                "S5",
                "S6",
                "S7",
                "S8",
                "S9",
                "S10",
                "S11",
                "S12",
            ],
            "group": ["A", "B", "C", "A", "B", "C", "A", "B", "C", "A", "B", "C"],
            "protein": [
                1.4,
                6.2,
                9.03,
                2.3,
                7.4,
                14.01,
                4.5,
                9.6,
                10.4,
                7.5,
                11.6,
                16.4,
            ],
            "group2": ["a", "a", "a", "a", "a", "a", "b", "b", "b", "b", "b", "b"],
        }
        self.df = pd.DataFrame(self.data)

    def test_ttest_ind(self):
        data = {
            "group1": [1.3, 1.2, 0.2, 2.2, 3.3, 8.43],
            "group2": [1.4, 6.2, 7.3, 0.4, 1.5, 0.6],
        }
        df = pd.DataFrame(data)

        condition1 = "group1"
        condition2 = "group2"
        is_logged = True
        non_par = False

        expected_result = pg.ttest(df[condition1], df[condition2], paired=False)
        result = dr.calculate_ttest(
            df, condition1, condition2, is_logged=is_logged, non_par=non_par
        )

        self.assertAlmostEqual(result[0], expected_result["T"].values[0])
        self.assertAlmostEqual(result[1], expected_result["p-val"].values[0])

    def test_mann_whitney(self):
        data = {"group1": [1, 2, 3, 4, 5], "group2": [2, 4, 6, 8, 10]}
        df = pd.DataFrame(data)

        condition1 = "group1"
        condition2 = "group2"
        is_logged = True
        non_par = True

        expected_result = pg.mwu(df[condition1], df[condition2])
        result = dr.calculate_ttest(
            df, condition1, condition2, is_logged=is_logged, non_par=non_par
        )

        self.assertAlmostEqual(result[0], expected_result["U-val"].values[0])
        self.assertAlmostEqual(result[1], expected_result["p-val"].values[0])

    def test_calculate_anova(self):
        column = "protein"
        group = "group"

        expected_result = pg.anova(data=self.df, dv=column, between=group)
        expected_t, expected_df1, expected_df2, expected_pvalue = expected_result[
            ["F", "ddof1", "ddof2", "p-unc"]
        ].values[0]

        result = dr.calculate_anova(self.df, column, group=group)

        self.assertEqual(result[1], expected_df1)
        self.assertEqual(result[2], expected_df2)
        self.assertEqual(result[3], expected_t)
        self.assertEqual(result[4], expected_pvalue)

    def test_calculate_ancova(self):
        column = "protein"
        group = "group"
        covariates = []

        expected_result = pg.ancova(
            data=self.df, dv=column, between=group, covar=covariates
        )
        expected_t, expected_df, expected_pvalue = expected_result.loc[
            expected_result["Source"] == group, ["F", "DF", "p-unc"]
        ].values[0]

        result = dr.calculate_ancova(
            self.df, column, group=group, covariates=covariates
        )

        self.assertEqual(result[1], expected_df)
        self.assertEqual(result[2], expected_df)
        self.assertEqual(result[3], expected_t)
        self.assertEqual(result[4], expected_pvalue)

    def test_calculate_repeated_measures_anova(self):
        """Source     SS  DF      MS         F     p-unc       ng2  eps
        0  group   1.50   1   1.500  0.087642  0.795106  0.032609  1.0
        1  Error  34.23   2  17.115       NaN       NaN       NaN  NaN"""
        column = "protein"
        subject = "subject"
        within = "group"

        expected_result = pg.rm_anova(
            data=self.df,
            dv=column,
            within=within,
            subject=subject,
            detailed=True,
            correction=True,
        )
        expected_t, expected_pvalue = expected_result.loc[
            0, ["F", "p-unc"]
        ].values.tolist()
        expected_df1, expected_df2 = expected_result["DF"]
        result = dr.calculate_repeated_measures_anova(
            self.df, column, subject=subject, within=within
        )

        self.assertEqual(result[1], expected_df1)
        self.assertEqual(result[2], expected_df2)
        self.assertEqual(result[3], expected_t)
        self.assertEqual(result[4], expected_pvalue)

    def test_calculate_mixed_anova(self):
        column = "protein"
        subject = "subject"
        within = "group"
        between = "group2"

        expected_result = pg.mixed_anova(
            data=self.df,
            dv=column,
            within=within,
            between=between,
            subject=subject,
            correction=True,
        )
        expected_result["identifier"] = column
        expected_result = expected_result[
            ["identifier", "DF1", "DF2", "F", "p-unc", "Source"]
        ]

        result = dr.calculate_mixed_anova(
            self.df, column, subject=subject, within=within, between=between
        )

        self.assertEqual(result.values.tolist(), expected_result.values.tolist())


if __name__ == "__main__":
    unittest.main()



"""Uniprot API user functions for fetching annotations for UniProt IDs and providing 
the results as a pandas.DataFrame."""

import pandas as pd

from .uniprot import (
    check_id_mapping_results_ready,
    get_id_mapping_results_link,
    get_id_mapping_results_search,
    submit_id_mapping,
)


# function for outside usage
def fetch_annotations(
    ids: pd.Index | list,
    fields: str = "accession,go_p,go_c,go_f",
) -> pd.DataFrame:
    """Fetch annotations for UniProt IDs. Combines several calls to the API of UniProt's
    knowledgebase (KB).

    Parameters
    ----------
    ids : pd.Index | list
        Iterable of UniProt IDs. Fetches annotations as speecified by the specified fields.
    fields : str, optional
        Fields to fetch, by default "accession,go_p,go_c. See for availble fields:
        https://www.uniprot.org/help/return_fields

    Returns
    -------
    pd.DataFrame
        DataFrame with annotations of the UniProt IDs.
    """
    job_id = submit_id_mapping(from_db="UniProtKB_AC-ID", to_db="UniProtKB", ids=ids)
    # tsv used here to return a DataFrame. Maybe other formats are availale at some points
    _format = "tsv"
    if check_id_mapping_results_ready(job_id):
        link = get_id_mapping_results_link(job_id)
        # add fields to the link to get more information
        # From and Entry (accession) are the same for UniProt IDs.
        results = get_id_mapping_results_search(
            link + f"?fields={fields}&format={_format}"
        )
    header = results.pop(0).split("\t")
    results = [line.split("\t") for line in results]
    df = pd.DataFrame(results, columns=header)
    return df


def process_annotations(annotations: pd.DataFrame, fields: str) -> pd.DataFrame:
    """Process annotations fetched from UniProt API.

    Parameters
    ----------
    annotations : pd.DataFrame
        DataFrame with annotations fetched from UniProt API.
    fields : str
        Fields that were fetched from the API. Comma-separated string. Fields
        needs to match number of columns in annotations.

    Returns
    -------
    pd.DataFrame
        Processed DataFrame with annotations in long-format.
    """
    d_fields_to_col = {
        k: v for k, v in zip(fields.split(","), annotations.columns[1:], strict=True)
    }

    # expand go terms
    to_expand = list()
    for field in d_fields_to_col:
        if "go_" in field:
            col = d_fields_to_col[field]
            annotations[col] = annotations[col].str.split(";")
            to_expand.append(col)
    for col in to_expand:
        # this is a bit wastefull. Processing to stack format should be done here.
        annotations = annotations.explode(col, ignore_index=True)
    # process other than go term columns
    annotations = (
        annotations.set_index("From")
        .rename_axis("identifier")
        # .drop("Entry", axis=1)
        .rename_axis("source", axis=1)
        .stack()
        .to_frame("annotation")
        .reset_index()
        .drop_duplicates(ignore_index=True)
    )
    return annotations
